from __future__ import annotations

import asyncio
import logging

import zigpy.zdo
import zigpy.device
import zigpy.zdo.types as zdo_t
import zigpy.application

import zigpy_znp.types as t
import zigpy_znp.commands as c
import zigpy_znp.zigbee.application as znp_app

LOGGER = logging.getLogger(__name__)

NWK_UPDATE_LOOP_DELAY = 1


class ZNPCoordinator(zigpy.device.Device):
    """
    Coordinator zigpy device that keeps track of our endpoints and clusters.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        assert hasattr(self, "zdo")
        self.zdo = ZNPZDOEndpoint(self)
        self.endpoints[0] = self.zdo

    @property
    def manufacturer(self):
        return "Texas Instruments"

    @property
    def model(self):
        if self.application._znp.version > 3.0:
            model = "CC1352/CC2652"
            version = "3.30+"
        else:
            model = "CC2538" if self.application._znp.nvram.align_structs else "CC2531"
            version = "Home 1.2" if self.application._znp.version == 1.2 else "3.0.x"

        return f"{model}, Z-Stack {version} (build {self.application._zstack_build_id})"

    def request(
        self,
        profile,
        cluster,
        src_ep,
        dst_ep,
        sequence,
        data,
        expect_reply=True,
        # Extend the default timeout
        timeout=2 * zigpy.device.APS_REPLY_TIMEOUT,
        use_ieee=False,
    ):
        """
        Normal `zigpy.device.Device:request` except its default timeout is longer.
        """

        return super().request(
            profile,
            cluster,
            src_ep,
            dst_ep,
            sequence,
            data,
            expect_reply=expect_reply,
            timeout=timeout,
            use_ieee=use_ieee,
        )


class ZNPZDOEndpoint(zigpy.zdo.ZDO):
    @property
    def app(self) -> zigpy.application.ControllerApplication:
        return self.device.application

    def _send_loopback_reply(
        self, command_id: zdo_t.ZDOCmd, *, tsn: t.uint8_t, **kwargs
    ):
        """
        Constructs and sends back a loopback ZDO response.
        """

        message = t.uint8_t(tsn).serialize() + self._serialize(
            command_id, *kwargs.values()
        )

        LOGGER.debug("Sending loopback reply %s (%s), tsn=%s", command_id, kwargs, tsn)

        self.app.handle_message(
            sender=self.app.zigpy_device,
            profile=znp_app.ZDO_PROFILE,
            cluster=command_id,
            src_ep=znp_app.ZDO_ENDPOINT,
            dst_ep=znp_app.ZDO_ENDPOINT,
            message=message,
        )

    def handle_mgmt_nwk_update_req(
        self, hdr: zdo_t.ZDOHeader, NwkUpdate: zdo_t.NwkUpdate, *, dst_addressing
    ):
        """
        Handles ZDO `Mgmt_NWK_Update_req` sent to the coordinator.
        """

        self.create_catching_task(
            self.async_handle_mgmt_nwk_update_req(
                hdr, NwkUpdate, dst_addressing=dst_addressing
            )
        )

    async def async_handle_mgmt_nwk_update_req(
        self, hdr: zdo_t.ZDOHeader, NwkUpdate: zdo_t.NwkUpdate, *, dst_addressing
    ):
        # Energy scans are handled properly by Z-Stack, no need to do anything
        if NwkUpdate.ScanDuration not in (
            zdo_t.NwkUpdate.CHANNEL_CHANGE_REQ,
            zdo_t.NwkUpdate.CHANNEL_MASK_MANAGER_ADDR_CHANGE_REQ,
        ):
            return

        old_network_info = self.app.state.network_information

        if (
            t.Channels.from_channel_list([old_network_info.channel])
            == NwkUpdate.ScanChannels
        ):
            LOGGER.warning("NWK update request is ignored when channel does not change")
            self._send_loopback_reply(
                zdo_t.ZDOCmd.Mgmt_NWK_Update_rsp,
                Status=zdo_t.Status.SUCCESS,
                ScannedChannels=t.Channels.NO_CHANNELS,
                TotalTransmissions=0,
                TransmissionFailures=0,
                EnergyValues=[],
                tsn=hdr.tsn,
            )
            return

        await self.app._znp.request(
            request=c.ZDO.MgmtNWKUpdateReq.Req(
                Dst=0x0000,
                DstAddrMode=t.AddrMode.NWK,
                Channels=NwkUpdate.ScanChannels,
                ScanDuration=NwkUpdate.ScanDuration,
                # Missing fields in the request cannot be `None` in the Z-Stack command
                ScanCount=NwkUpdate.ScanCount or 0,
                NwkManagerAddr=NwkUpdate.nwkManagerAddr or 0x0000,
            ),
            RspStatus=t.Status.SUCCESS,
        )

        # Wait until the network info changes, it can take ~5s
        while (
            self.app.state.network_information.nwk_update_id
            == old_network_info.nwk_update_id
        ):
            await self.app.load_network_info(load_devices=False)
            await asyncio.sleep(NWK_UPDATE_LOOP_DELAY)

        # Z-Stack automatically increments the NWK update ID instead of setting it
        # TODO: Directly set it once radio settings API is finalized.
        if NwkUpdate.nwkUpdateId != self.app.state.network_information.nwk_update_id:
            LOGGER.warning(
                f"`nwkUpdateId` was incremented to"
                f" {self.app.state.network_information.nwk_update_id} instead of being"
                f" set to {NwkUpdate.nwkUpdateId}"
            )

        self._send_loopback_reply(
            zdo_t.ZDOCmd.Mgmt_NWK_Update_rsp,
            Status=zdo_t.Status.SUCCESS,
            ScannedChannels=t.Channels.NO_CHANNELS,
            TotalTransmissions=0,
            TransmissionFailures=0,
            EnergyValues=[],
            tsn=hdr.tsn,
        )



import typing
import asyncio
import logging
import warnings

import serial

import zigpy_znp.config as conf
import zigpy_znp.frames as frames
import zigpy_znp.logger as log
from zigpy_znp.types import Bytes
from zigpy_znp.exceptions import InvalidFrame

with warnings.catch_warnings():
    warnings.filterwarnings(
        action="ignore",
        module="serial_asyncio",
        message='"@coroutine" decorator is deprecated',
        category=DeprecationWarning,
    )
    import serial_asyncio  # noqa: E402


LOGGER = logging.getLogger(__name__)


class BufferTooShort(Exception):
    pass


class ZnpMtProtocol(asyncio.Protocol):
    def __init__(self, api):
        self._buffer = bytearray()
        self._api = api
        self._transport = None
        self._connected_event = asyncio.Event()

    def close(self) -> None:
        """Closes the port."""

        self._api = None
        self._buffer.clear()

        if self._transport is not None:
            LOGGER.debug("Closing serial port")

            self._transport.close()
            self._transport = None

    def connection_lost(self, exc: typing.Optional[Exception]) -> None:
        """Connection lost."""

        if exc is not None:
            LOGGER.warning("Lost connection", exc_info=exc)

        if self._api is not None:
            self._api.connection_lost(exc)

    def connection_made(self, transport: serial_asyncio.SerialTransport) -> None:
        """Opened serial port."""
        self._transport = transport
        LOGGER.debug("Opened %s serial port", transport.serial.name)

        self._connected_event.set()

        if self._api is not None:
            self._api.connection_made()

    def data_received(self, data: bytes) -> None:
        """Callback when data is received."""
        self._buffer += data

        LOGGER.log(log.TRACE, "Received data: %s", Bytes.__repr__(data))

        for frame in self._extract_frames():
            LOGGER.log(log.TRACE, "Parsed frame: %s", frame)

            try:
                self._api.frame_received(frame.payload)
            except Exception as e:
                LOGGER.error(
                    "Received an exception while passing frame to API: %s",
                    frame,
                    exc_info=e,
                )

    def send(self, payload: frames.GeneralFrame) -> None:
        """Sends data taking care of framing."""
        self.write(frames.TransportFrame(payload).serialize())

    def write(self, data: bytes) -> None:
        """
        Writes raw bytes to the transport. This method should be used instead of
        directly writing to the transport with `transport.write`.
        """

        LOGGER.log(log.TRACE, "Sending data: %s", Bytes.__repr__(data))
        self._transport.write(data)

    def set_dtr_rts(self, *, dtr: bool, rts: bool) -> None:
        LOGGER.debug("Setting serial pin states: DTR=%s, RTS=%s", dtr, rts)

        self._transport.serial.dtr = dtr
        self._transport.serial.rts = rts

    @property
    def name(self) -> str:
        return self._transport.serial.name

    @property
    def baudrate(self) -> int:
        return self._transport.serial.baudrate

    def _extract_frames(self) -> typing.Iterator[frames.TransportFrame]:
        """Extracts frames from the buffer until it is exhausted."""
        while True:
            try:
                yield self._extract_frame()
            except BufferTooShort:
                # If the buffer is too short, there is nothing more we can do
                break
            except InvalidFrame:
                # If the buffer contains invalid data, drop it until we find the SoF
                sof_index = self._buffer.find(frames.TransportFrame.SOF, 1)

                if sof_index < 0:
                    # If we don't have a SoF in the buffer, drop everything
                    self._buffer.clear()
                else:
                    del self._buffer[:sof_index]

    def _extract_frame(self) -> frames.TransportFrame:
        """Extracts a single frame from the buffer."""

        # The shortest possible frame is 5 bytes long
        if len(self._buffer) < 5:
            raise BufferTooShort()

        # The buffer must start with a SoF
        if self._buffer[0] != frames.TransportFrame.SOF:
            raise InvalidFrame()

        length = self._buffer[1]

        # If the packet length field exceeds 250, our packet is not valid
        if length > 250:
            raise InvalidFrame()

        # Don't bother deserializing anything if the packet is too short
        # [SoF:1] [Length:1] [Command:2] [Data:(Length)] [FCS:1]
        if len(self._buffer) < length + 5:
            raise BufferTooShort()

        # At this point we should have a complete frame
        # If not, deserialization will fail and the error will propapate up
        frame, rest = frames.TransportFrame.deserialize(self._buffer)

        # If we get this far then we have a valid frame. Update the buffer.
        del self._buffer[: len(self._buffer) - len(rest)]

        return frame

    def __repr__(self) -> str:
        return (
            f"<"
            f"{type(self).__name__} connected to {self.name!r}"
            f" at {self.baudrate} baud"
            f" (api: {self._api})"
            f">"
        )


async def connect(config: conf.ConfigType, api) -> ZnpMtProtocol:
    loop = asyncio.get_running_loop()

    port = config[conf.CONF_DEVICE_PATH]
    baudrate = config[conf.CONF_DEVICE_BAUDRATE]
    flow_control = config[conf.CONF_DEVICE_FLOW_CONTROL]

    LOGGER.debug("Connecting to %s at %s baud", port, baudrate)

    _, protocol = await serial_asyncio.create_serial_connection(
        loop=loop,
        protocol_factory=lambda: ZnpMtProtocol(api),
        url=port,
        baudrate=baudrate,
        parity=serial.PARITY_NONE,
        stopbits=serial.STOPBITS_ONE,
        xonxoff=(flow_control == "software"),
        rtscts=(flow_control == "hardware"),
    )

    await protocol._connected_event.wait()

    LOGGER.debug("Connected to %s at %s baud", port, baudrate)

    return protocol



from __future__ import annotations

import functools
import dataclasses

import zigpy_znp.types as t
from zigpy_znp.exceptions import InvalidFrame


@dataclasses.dataclass(frozen=True)
class GeneralFrame:
    header: t.CommandHeader
    data: t.Bytes

    def __post_init__(self) -> None:
        # We're frozen so `self.header = ...` is disallowed
        if not isinstance(self.header, t.CommandHeader):
            object.__setattr__(self, "header", t.CommandHeader(self.header))

        if not isinstance(self.data, t.Bytes):
            object.__setattr__(self, "data", t.Bytes(self.data))

        if self.length > 250:
            raise InvalidFrame(
                f"Frame length cannot exceed 250 bytes. Got: {self.length}"
            )

    @property
    def length(self) -> t.uint8_t:
        """Length of the frame."""
        return t.uint8_t(len(self.data))

    @classmethod
    def deserialize(cls, data):
        """Deserialize frame and sanity checks."""
        length, data = t.uint8_t.deserialize(data)

        if length > 250:
            raise InvalidFrame(f"Frame length cannot exceed 250 bytes. Got: {length}")

        if len(data) < length + 2:
            raise InvalidFrame(f"Data is too short for {cls.__name__}")

        header, data = t.CommandHeader.deserialize(data)
        payload, data = data[:length], data[length:]
        return cls(header, payload), data

    def serialize(self) -> bytes:
        return self.length.serialize() + self.header.serialize() + self.data.serialize()


@dataclasses.dataclass
class TransportFrame:
    """Transport frame."""

    SOF = t.uint8_t(0xFE)  # Start of frame marker

    payload: GeneralFrame

    @classmethod
    def deserialize(cls, data: bytes) -> tuple[TransportFrame, bytes]:
        sof, data = t.uint8_t.deserialize(data)

        if sof != cls.SOF:
            raise InvalidFrame(
                f"Expected frame to start with SOF 0x{cls.SOF:02X}, got 0x{sof:02X}"
            )

        gen_frame, data = GeneralFrame.deserialize(data)
        checksum, data = t.uint8_t.deserialize(data)

        frame = cls(gen_frame)

        if frame.checksum() != checksum:
            raise InvalidFrame(
                f"Invalid frame checksum for data {gen_frame}: "
                f"expected 0x{frame.checksum():02X}, got 0x{checksum:02X}"
            )

        return frame, data

    def checksum(self) -> t.uint8_t:
        """
        Calculates the FCS of the payload.
        """

        checksum = functools.reduce(lambda a, b: a ^ b, self.payload.serialize())
        return t.uint8_t(checksum)

    def serialize(self) -> bytes:
        return (
            self.SOF.serialize()
            + self.payload.serialize()
            + self.checksum().serialize()
        )



import sys
import asyncio
import logging

from zigpy_znp.tools.common import setup_parser
from zigpy_znp.zigbee.application import ControllerApplication

LOGGER = logging.getLogger(__name__)


async def form_network(radio_path):
    LOGGER.info("Starting up zigpy-znp")

    config = ControllerApplication.SCHEMA({"device": {"path": radio_path}})
    app = ControllerApplication(config)

    await app.startup(force_form=True)
    await app.shutdown()


async def main(argv):
    parser = setup_parser("Form a network with randomly generated settings")
    args = parser.parse_args(argv)

    await form_network(args.serial)


if __name__ == "__main__":
    asyncio.run(main(sys.argv[1:]))  # pragma: no cover



import typing
import numbers

import voluptuous as vol
from zigpy.config import (  # noqa: F401
    CONF_NWK,
    CONF_DEVICE,
    CONF_NWK_KEY,
    CONFIG_SCHEMA,
    SCHEMA_DEVICE,
    CONF_NWK_PAN_ID,
    CONF_DEVICE_PATH,
    CONF_NWK_CHANNEL,
    CONF_NWK_CHANNELS,
    CONF_NWK_UPDATE_ID,
    CONF_NWK_TC_ADDRESS,
    CONF_NWK_TC_LINK_KEY,
    CONF_NWK_EXTENDED_PAN_ID,
    cv_boolean,
)

from zigpy_znp.commands.util import LEDMode

ConfigType = typing.Dict[str, typing.Any]

VolPositiveNumber = vol.All(numbers.Real, vol.Range(min=0))

CONF_DEVICE_BAUDRATE = "baudrate"
CONF_DEVICE_FLOW_CONTROL = "flow_control"

SCHEMA_DEVICE = SCHEMA_DEVICE.extend(
    {
        vol.Optional(CONF_DEVICE_BAUDRATE, default=115_200): int,
        vol.Optional(CONF_DEVICE_FLOW_CONTROL, default=None): vol.In(
            ("hardware", "software", None)
        ),
    }
)


def EnumValue(enum, transformer=str):
    def validator(v):
        if isinstance(v, enum):
            return v

        return enum[transformer(v)]

    return validator


def keys_have_same_length(*keys):
    def validator(config):
        lengths = [len(config[k]) for k in keys]

        if len(set(lengths)) != 1:
            raise vol.Invalid(
                f"Values for {keys} must all have the same length: {lengths}"
            )

        return config

    return validator


def bool_to_upper_str(value: typing.Any) -> str:
    """
    Converts the value into an uppercase string, including unquoted YAML booleans.
    """

    if value is True:
        return "ON"
    elif value is False:
        return "OFF"
    else:
        return str(value).upper()


CONF_ZNP_CONFIG = "znp_config"
CONF_TX_POWER = "tx_power"
CONF_LED_MODE = "led_mode"
CONF_SKIP_BOOTLOADER = "skip_bootloader"
CONF_SREQ_TIMEOUT = "sync_request_timeout"
CONF_ARSP_TIMEOUT = "async_response_timeout"
CONF_AUTO_RECONNECT_RETRY_DELAY = "auto_reconnect_retry_delay"
CONF_MAX_CONCURRENT_REQUESTS = "max_concurrent_requests"
CONF_CONNECT_RTS_STATES = "connect_rts_pin_states"
CONF_CONNECT_DTR_STATES = "connect_dtr_pin_states"

CONFIG_SCHEMA = CONFIG_SCHEMA.extend(
    {
        vol.Required(CONF_DEVICE): SCHEMA_DEVICE,
        vol.Optional(CONF_ZNP_CONFIG, default={}): vol.Schema(
            vol.All(
                {
                    vol.Optional(CONF_TX_POWER, default=None): vol.Any(
                        None, vol.All(int, vol.Range(min=-22, max=22))
                    ),
                    vol.Optional(CONF_SREQ_TIMEOUT, default=15): VolPositiveNumber,
                    vol.Optional(CONF_ARSP_TIMEOUT, default=30): VolPositiveNumber,
                    vol.Optional(
                        CONF_AUTO_RECONNECT_RETRY_DELAY, default=5
                    ): VolPositiveNumber,
                    vol.Optional(CONF_SKIP_BOOTLOADER, default=True): cv_boolean,
                    vol.Optional(CONF_LED_MODE, default=LEDMode.OFF): vol.Any(
                        None, EnumValue(LEDMode, transformer=bool_to_upper_str)
                    ),
                    vol.Optional(CONF_MAX_CONCURRENT_REQUESTS, default="auto"): vol.Any(
                        "auto", VolPositiveNumber
                    ),
                    vol.Optional(
                        CONF_CONNECT_RTS_STATES, default=[False, True, False]
                    ): vol.Schema([cv_boolean]),
                    vol.Optional(
                        CONF_CONNECT_DTR_STATES, default=[False, False, False]
                    ): vol.Schema([cv_boolean]),
                },
                keys_have_same_length(CONF_CONNECT_RTS_STATES, CONF_CONNECT_DTR_STATES),
            )
        ),
    }
)



import sys
import asyncio
import logging

import async_timeout

import zigpy_znp.commands as c
from zigpy_znp.api import ZNP
from zigpy_znp.config import CONFIG_SCHEMA
from zigpy_znp.tools.common import ClosableFileType, setup_parser

LOGGER = logging.getLogger(__name__)


async def read_firmware(znp: ZNP) -> bytearray:
    try:
        async with async_timeout.timeout(5):
            handshake_rsp = await znp.request_callback_rsp(
                request=c.UBL.HandshakeReq.Req(),
                callback=c.UBL.HandshakeRsp.Callback(partial=True),
            )
    except asyncio.TimeoutError:
        raise RuntimeError(
            "Did not receive a bootloader handshake response!"
            " Make sure your adapter has just been plugged in and"
            " nothing else has had a chance to communicate with it. Alternatively,"
            " press the button furthest from the USB port. The LED should turn red."
        )

    if handshake_rsp.Status != c.ubl.BootloaderStatus.SUCCESS:
        raise RuntimeError(f"Bad bootloader handshake response: {handshake_rsp}")

    # All reads and writes are this size
    buffer_size = handshake_rsp.BufferSize

    data = bytearray()

    for offset in range(0, c.ubl.IMAGE_SIZE, buffer_size):
        address = offset // c.ubl.FLASH_WORD_SIZE
        LOGGER.info("Progress: %0.2f%%", (100.0 * offset) / c.ubl.IMAGE_SIZE)

        read_rsp = await znp.request_callback_rsp(
            request=c.UBL.ReadReq.Req(FlashWordAddr=address),
            callback=c.UBL.ReadRsp.Callback(partial=True),
        )

        assert read_rsp.Status == c.ubl.BootloaderStatus.SUCCESS
        assert read_rsp.FlashWordAddr == address
        assert len(read_rsp.Data) == buffer_size

        data.extend(read_rsp.Data)

    return data


async def main(argv):
    parser = setup_parser("Backup a radio's firmware")
    parser.add_argument(
        "--output",
        "-o",
        type=ClosableFileType("wb"),
        help="Output .bin file",
        required=True,
    )

    args = parser.parse_args(argv)

    with args.output as f:
        znp = ZNP(
            CONFIG_SCHEMA(
                {
                    "znp_config": {"skip_bootloader": False},
                    "device": {"path": args.serial},
                }
            )
        )

        # The bootloader handshake must be the very first command
        await znp.connect(test_port=False)

        data = await read_firmware(znp)
        znp.close()

        f.write(data)

        LOGGER.info("Unplug your adapter to leave bootloader mode!")


if __name__ == "__main__":
    asyncio.run(main(sys.argv[1:]))  # pragma: no cover
